#include <windows.h>
#include "../DLLMisc/hash.h"
#include "../DLLMisc/token.h"

#define VERSION "1.02"

#define AUTHINFO \
"Standard Input/Output/Error DLL for mIRC, v" VERSION " (c) 2000 DragonZap, dragonzap1@aol.com"

#ifdef __cplusplus
extern "C" {
#endif

#define BINBUF_SIZE 200
#define OUTBUF_SIZE 512

#define TP_OUT			0x01
#define TP_ERR			0x02
#define TP_OUT_TRUNCATE TP_OUT
#define TP_ERR_TRUNCATE TP_ERR
#define TP_KEEPALIVE	0x04

struct TEXTPROCESS {
	char fBlock;
	HANDLE hStdIn;
	HANDLE hStdInR;
	HANDLE hStdOut;
	HANDLE hStdErr;
	HANDLE hProcess;
	DWORD dwThreadId;
};
#define PROCSIZE 13

char Installed = 0;
HMODULE Self = NULL;
HASHTABLE *lpProcs = NULL;
/////////////////////////// ERROR STRINGS ////////////////////////////////
#define STD_ERROR "ERROR STDE"
#define STDE_SYNTAX			"1 Error in request"
#define STDE_NAME			"2 Process alias name is already in use"
#define STDE_NONAME			"3 Process alias name is not in use"
#define STDE_PROCESS		"4 Process could not be started"
#define STDE_WAIT			"5 No data ready for reading"
#define STDE_EOF			"6 End of file"
#define STDE_INPUT			"7 Input has already been closed"
#define STDE_ALIVE			"8 Process is still alive"
#define STDE_DEAD			"9 Process is dead"
#define STDE_BADNAME		"10 Invalid process alias name"
#define STDE_UNKNOWN		"255 Unknown error"
#define RETURNERROR(e) {\
	strcpy(data,STD_ERROR STDE_ ## e );\
	return 3;\
}

//////////////////////////////////////////////////////////////////////////

// It's not multithreaded, so it's ok to have
char szBuffer[OUTBUF_SIZE];

bool CloseProcHandles(LPHASHCHAIN lpChain,void *lpContext) {
	TEXTPROCESS * lpTP = (TEXTPROCESS *)lpChain->hashValue;
	if (lpTP->hStdIn)
	{
		CloseHandle(lpTP->hStdIn);
		CloseHandle(lpTP->hStdInR);
	}
	CloseHandle(lpTP->hStdOut);
	CloseHandle(lpTP->hStdErr);

	if (!(lpTP->fBlock & TP_KEEPALIVE))
	{
		// mIRC wants dead. Now.
		PostThreadMessage(lpTP->dwThreadId,WM_QUIT,0,0);
		Sleep(0);
		TerminateProcess(lpTP->hProcess,1);
	}
	CloseHandle(lpTP->hProcess);

	lpTP->hProcess = lpTP->hStdIn = lpTP->hStdInR = lpTP->hStdOut =
		lpTP->hStdErr = NULL;	
	return 0;
}

struct PROCSEARCH {
	char * procName;
	DWORD procNum;
	DWORD wantProcNum;
	TEXTPROCESS *lpTP;
};

bool SearchProcsByNum(LPHASHCHAIN lpChain,void *lpContext) {
	PROCSEARCH *ps = (PROCSEARCH *)lpContext;
	if (ps->procNum == ps->wantProcNum)
	{
		ps->lpTP = (TEXTPROCESS *)lpChain->hashValue;
		ps->procName = (char *)lpChain->hashKey;
		return 1;
	}
	ps->procNum++;
	return 0;
}

BOOL WINAPI DllMain(HINSTANCE hinstDLL,DWORD fdwReason,LPVOID lpvReserved)
{
	Self = hinstDLL;
	if (fdwReason == DLL_PROCESS_DETACH)
	{
		if (lpProcs)
		{
			Hash_Enum(lpProcs,CloseProcHandles,NULL);
			Hash_Delete(lpProcs,true);
		}
	}
	return 1;
}

void StayLoaded()
{
	if (!Installed)
	{
		lpProcs = CreateHashTable(PROCSIZE);

		char DLLName[MAX_PATH+1];
		GetModuleFileName(Self,DLLName,MAX_PATH);
		LoadLibrary(DLLName);
		Installed = 1;
	}
}


int __stdcall GetProcess(HWND mWnd, HWND aWnd, char *data, char *parms,
						 BOOL show, BOOL nopause)
{
	StayLoaded(); 

	char *lpszData = data + 3;
	TEXTPROCESS * lpTP;

	if (hasNumTok(data,' ',false))
	{
		DWORD whichProc = atoi(data);
		DWORD procCount = Hash_GetKeyCount(lpProcs);
		if (whichProc == 0)
		{
			strcpy(data,"OK ");
			itoa(procCount,data+3,10);
			return 3;
		}
		whichProc--;
		if (whichProc >= procCount)
			RETURNERROR(NONAME);
		PROCSEARCH ps;
		ps.procName = NULL;
		ps.procNum = 0;
		ps.wantProcNum = whichProc;
		Hash_Enum(lpProcs,SearchProcsByNum,&ps);
		if (ps.procName == NULL)
			RETURNERROR(NONAME);
		lpTP = ps.lpTP;
		strcpy(data,"OK ");
		strcpy(lpszData,ps.procName);
		lpszData += strlen(lpszData);

	}
	else if (!Hash_GetKeyCI(lpProcs,data,(void **)&lpTP))
		RETURNERROR(NONAME)
	else
	{
		// Move existing data 3 bytes forward so that "OK " will fit
		char *lpDataMove = data+strlen(data);
		lpszData = lpDataMove+3;
		while(lpDataMove >= data)
		{
			lpDataMove[3] = lpDataMove[0];
			lpDataMove--;
		}
		strncpy(data,"OK ",3);
	}
	// Check for input, output, then error
	DWORD bytesAvail;
	DWORD procStat;
	GetExitCodeProcess(lpTP->hProcess,&procStat);

	// If the process is dead, no input is available
	if ((lpTP->hStdIn == NULL) || (procStat != STILL_ACTIVE))
		bytesAvail = OUTBUF_SIZE;
	else
		if (!PeekNamedPipe(lpTP->hStdInR,szBuffer,OUTBUF_SIZE,&bytesAvail,NULL,NULL))
			return 3;
	*(lpszData++)=' ';
	itoa(OUTBUF_SIZE - bytesAvail,lpszData,10);
	lpszData += strlen(lpszData);
	if (!PeekNamedPipe(lpTP->hStdOut,szBuffer,OUTBUF_SIZE,NULL,&bytesAvail,NULL))
		return 3;
	*(lpszData++)=' ';
	itoa(bytesAvail,lpszData,10);
	lpszData += strlen(lpszData);
	if (!PeekNamedPipe(lpTP->hStdErr,szBuffer,OUTBUF_SIZE,NULL,&bytesAvail,NULL))
		return 3;
	*(lpszData++)=' ';
	itoa(bytesAvail,lpszData,10);
	return 3;
}

int __stdcall RunConsole(HWND mWnd, HWND aWnd, char *data, char *parms,
						 BOOL show, BOOL nopause)
{
	StayLoaded();

	char *lpszCmdLine = nextTok(data,' ',true);

	if (hasNumTok(data,' ',false))
		RETURNERROR(BADNAME);

	if (Hash_GetKeyCI(lpProcs,data,NULL))
		RETURNERROR(NAME);

	TEXTPROCESS * lpNewProcess = new TEXTPROCESS;
	if (!lpNewProcess)
		RETURNERROR(UNKNOWN);

	lpNewProcess->fBlock = 0;
	lpNewProcess->hProcess = NULL;
	lpNewProcess->hStdInR = NULL;
	lpNewProcess->hStdIn = NULL;
	lpNewProcess->hStdOut = NULL;
	lpNewProcess->hStdErr = NULL;

	HANDLE hStdOutW = NULL;
	HANDLE hStdErrW = NULL;
	
	SECURITY_ATTRIBUTES saPipe;
	saPipe.nLength = sizeof(SECURITY_ATTRIBUTES);
	saPipe.lpSecurityDescriptor = NULL;
	saPipe.bInheritHandle = true;

	if (!CreatePipe(&(lpNewProcess->hStdInR),&(lpNewProcess->hStdIn),
		&saPipe,OUTBUF_SIZE) ||
		!CreatePipe(&(lpNewProcess->hStdOut),&hStdOutW,&saPipe,OUTBUF_SIZE) ||
		!CreatePipe(&(lpNewProcess->hStdErr),&hStdErrW,&saPipe,OUTBUF_SIZE))
	{
		if (lpNewProcess->hStdIn)
			CloseHandle(lpNewProcess->hStdIn);
		if (lpNewProcess->hStdInR)
			CloseHandle(lpNewProcess->hStdInR);
		if (lpNewProcess->hStdOut)
			CloseHandle(lpNewProcess->hStdOut);
		if (hStdOutW)
			CloseHandle(hStdOutW);
		if (lpNewProcess->hStdErr)
			CloseHandle(lpNewProcess->hStdErr);
		if (hStdErrW)
			CloseHandle(hStdErrW);
		delete lpNewProcess;
		RETURNERROR(UNKNOWN);
	}
	
	STARTUPINFO sInfo;
	PROCESS_INFORMATION pInfo;
	memset(&sInfo,0,sizeof(sInfo));
	sInfo.cb = sizeof(sInfo);
	sInfo.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;

	sInfo.wShowWindow = SW_HIDE;
	sInfo.hStdInput = lpNewProcess->hStdInR;
	sInfo.hStdOutput = hStdOutW;
	sInfo.hStdError = hStdErrW;
	

	int MadeProcess = CreateProcess(NULL,lpszCmdLine,NULL,NULL,true,
		NORMAL_PRIORITY_CLASS,NULL,NULL,&sInfo,&pInfo);
	
	CloseHandle(hStdOutW);
	CloseHandle(hStdErrW);
	
	if (!MadeProcess)
	{
		CloseHandle(lpNewProcess->hStdInR);
		CloseHandle(lpNewProcess->hStdIn);
		CloseHandle(lpNewProcess->hStdOut);
		CloseHandle(lpNewProcess->hStdErr);
		delete lpNewProcess;
		RETURNERROR(PROCESS);
	}

	// We don't need the thread's handle
	CloseHandle(pInfo.hThread);
	lpNewProcess->hProcess = pInfo.hProcess;
	lpNewProcess->dwThreadId = pInfo.dwThreadId;

	Hash_SetKeyCI(lpProcs,data,lpNewProcess);
	
	strcpy(data,"OK");
	return 3;
}




#define RL_OK	0
#define RL_EOF	1
#define RL_WAIT	2
#define RL_ERR  3
// Flushes the remainder of a line of text from the handle
char FlushLine(HANDLE hRead)
{
	// We have to clear out the garbage from the last giant read
	bool noLF = true;
	DWORD br;
	while(noLF)
	{
		if (!PeekNamedPipe(hRead,szBuffer,OUTBUF_SIZE,&br,NULL,NULL))
			return RL_ERR;
		if (br == 0)
			return RL_WAIT;
		DWORD cbyte;
		for(cbyte=0;cbyte<br;cbyte++)
			if (szBuffer[cbyte] == '\n')
			{
				noLF = false;
				break;
			}
		if (!ReadFile(hRead,szBuffer,(noLF ? OUTBUF_SIZE : cbyte+1),&br,NULL))
			return RL_ERR;
	}
	return RL_OK;
}

// lpszBuffer must have at least OUTBUF_SIZE+1 bytes
char ReadLine(TEXTPROCESS *lpReadProc,char WhichHandle,char * lpszBuffer)
{
	HANDLE hRead;
	switch(WhichHandle)
	{
	case TP_OUT:
		hRead = lpReadProc->hStdOut;
		break;
	case TP_ERR:
		hRead = lpReadProc->hStdErr;
		break;
	default:
		return RL_ERR;
	}

	// Figure out if the pipe has a new line to be read
	
	DWORD br;
	if (lpReadProc->fBlock & WhichHandle)
	{
		char FlushStat = FlushLine(hRead);
		if (FlushStat != RL_OK)
			return FlushStat;
		lpReadProc->fBlock &= ~WhichHandle;
	}
	if (!PeekNamedPipe(hRead,szBuffer,OUTBUF_SIZE,&br,NULL,NULL))
	{
		switch(GetLastError())
		{
		case ERROR_BROKEN_PIPE:
		case ERROR_NO_DATA:
		case ERROR_PIPE_NOT_CONNECTED:
			return RL_EOF;
		case ERROR_PIPE_LISTENING:
			return RL_WAIT;
		default:
			return RL_ERR;
		}
	}

	DWORD cbyte;
	for(cbyte=0;cbyte<br;cbyte++)
		if (szBuffer[cbyte] == '\n')
			break;
	if (cbyte >= OUTBUF_SIZE)
		// Read *everything*
	{
		lpReadProc->fBlock |= WhichHandle;
		cbyte = -1;
	}
	else if (cbyte >= br)
	{
		// Maybe the process has terminated?
		DWORD termStat;
		GetExitCodeProcess(lpReadProc->hProcess,&termStat);
		if (termStat == STILL_ACTIVE)
			// Waiting still
			return RL_WAIT;
		else
		{
			if (cbyte == 0)
				return RL_EOF;

			// The process is dead, but no \n! Therefore, read the rest anyway
		}
	}
	
	if (!ReadFile(hRead,szBuffer,(cbyte == -1 ? OUTBUF_SIZE : cbyte+1),&br,NULL))
		return RL_ERR;

	// Copy the data over, minus the last LF, and omitting all CRs
	DWORD newBufPos = 0;
	DWORD oldBufPos;
	for(oldBufPos = 0;oldBufPos<br;oldBufPos++)
		if (szBuffer[oldBufPos] != '\r')
			lpszBuffer[newBufPos++] = szBuffer[oldBufPos];

	if ((lpszBuffer[newBufPos-1] != '\n') || (newBufPos == 0))
		lpszBuffer[newBufPos] = 0;
	else
		lpszBuffer[newBufPos-1]=0;
	return RL_OK;
}

// TODO: Add option for reading as text
int __stdcall ReadBinary(HWND mWnd, HWND aWnd, char *data, char *parms,
						 BOOL show, BOOL nopause)
{
	StayLoaded();

	char *lpszWhichOut = nextTok(data,' ',true);
	char *lpszByteCount = nextTok(lpszWhichOut,' ',true);
	char *lpszOpts = nextTok(lpszByteCount,' ',true);
	
	TEXTPROCESS *lpTP;
	if (!Hash_GetKeyCI(lpProcs,data,(void **)&lpTP))
		RETURNERROR(NONAME);

	char binBuffer[BINBUF_SIZE];
	DWORD br;
	char WhichHandle;
	HANDLE hRead;
	if (!stricmp(lpszWhichOut,"out"))
	{
		hRead = lpTP->hStdOut;
		WhichHandle = TP_OUT;
	}
	else if (!stricmp(lpszWhichOut,"err"))
	{
		hRead = lpTP->hStdErr;
		WhichHandle = TP_ERR;
	}
	else
		RETURNERROR(SYNTAX);

	DWORD byteCount = atoi(lpszByteCount);
	if (byteCount < 0)
		RETURNERROR(SYNTAX);

	if (lpTP->fBlock & WhichHandle)
	{
		char FlushStat = FlushLine(hRead);
		if (FlushStat != RL_OK)
			return FlushStat;
		lpTP->fBlock &= ~WhichHandle;
	}

	if (!PeekNamedPipe(hRead,binBuffer,BINBUF_SIZE,&br,NULL,NULL))
	{
		switch(GetLastError())
		{
		case ERROR_BROKEN_PIPE:
		case ERROR_NO_DATA:
		case ERROR_PIPE_NOT_CONNECTED:
			return RL_EOF;
		case ERROR_PIPE_LISTENING:
			return RL_WAIT;
		default:
			return RL_ERR;
		}
	}

	// Max bytes allowed is br
	if (br == 0)
	{
		// Could either be a wait or an EOF
		DWORD termStat;
		GetExitCodeProcess(lpTP->hProcess,&termStat);
		if (termStat == STILL_ACTIVE)
			// Waiting still
			RETURNERROR(WAIT)
		else
			RETURNERROR(EOF);
	}

	if (byteCount > br)
		byteCount = br;

	if (!ReadFile(hRead,binBuffer,byteCount,&br,NULL))
		RETURNERROR(UNKNOWN);
	strcpy(data,"OK ");

	if (!stricmp(lpszOpts,"text"))
	{
		strncpy(data+3,binBuffer,br);
		data[3+br]=0;
	}
	else
	{
		char *lpszDPtr = data+3;
		for(byteCount = 0;byteCount < br;byteCount++)
		{
			itoa(binBuffer[byteCount],lpszDPtr,10);
			if (byteCount != br - 1)
				strcat(lpszDPtr," ");
			lpszDPtr += strlen(lpszDPtr);
		}
	}
	return 3;
}

int __stdcall ReadText(HWND mWnd, HWND aWnd, char *data, char *parms,
					   BOOL show, BOOL nopause)
{
	StayLoaded();

	char *lpszWhichOut = nextTok(data,' ',true);
	
	TEXTPROCESS *lpTP;
	if (!Hash_GetKeyCI(lpProcs,data,(void **)&lpTP))
		RETURNERROR(NONAME);

	char WhichOut;
	if (!stricmp(lpszWhichOut,"out"))
		WhichOut = TP_OUT;
	else if (!stricmp(lpszWhichOut,"err"))
		WhichOut = TP_ERR;
	else
		RETURNERROR(SYNTAX);
	strcpy(data,"OK ");
	char rStat = ReadLine(lpTP,WhichOut,data+3);
	switch(rStat)
	{
	case RL_OK:
		break;
	case RL_WAIT:
		RETURNERROR(WAIT);
	case RL_EOF:
		RETURNERROR(EOF);
	default:
		RETURNERROR(UNKNOWN);
	}
	return 3;

}

int __stdcall WriteText(HWND mWnd, HWND aWnd, char *data, char *parms,
						BOOL show, BOOL nopause)
{
	StayLoaded();

	char *lpszText = nextTok(data,' ',true);
	
	TEXTPROCESS *lpTP;
	if (!Hash_GetKeyCI(lpProcs,data,(void **)&lpTP))
		RETURNERROR(NONAME);

	DWORD procStat;
	GetExitCodeProcess(lpTP->hProcess,&procStat);
	if (procStat != STILL_ACTIVE)
		RETURNERROR(DEAD);

	if (!lpTP->hStdIn)
		RETURNERROR(INPUT);

	DWORD bw;
	DWORD textLen = strlen(lpszText);

	// Figure out how many bytes we CAN write
	PeekNamedPipe(lpTP->hStdInR,szBuffer,OUTBUF_SIZE,&bw,NULL,NULL);

	bw = OUTBUF_SIZE -bw;
	if (bw > textLen)
		bw = textLen;
	if (!WriteFile(lpTP->hStdIn,lpszText,bw,&bw,NULL))
		RETURNERROR(UNKNOWN);
	strcpy(data,"OK ");
	char *lpszNum = data+3;
	itoa(bw,lpszNum,10);
	lpszNum += strlen(lpszNum);
	*(lpszNum++) = ' ';
	itoa(textLen,lpszNum,10);
	return 3;
}

int __stdcall WriteBinary(HWND mWnd, HWND aWnd, char *data, char *parms,
						  BOOL show, BOOL nopause)
{
	StayLoaded();

	char *lpszBytes = nextTok(data,' ',true);
	
	TEXTPROCESS *lpTP;
	if (!Hash_GetKeyCI(lpProcs,data,(void **)&lpTP))
		RETURNERROR(NONAME);

	DWORD procStat;
	GetExitCodeProcess(lpTP->hProcess,&procStat);
	if (procStat != STILL_ACTIVE)
		RETURNERROR(DEAD);

	if (!lpTP->hStdIn)
		RETURNERROR(INPUT);

	char binBuffer[OUTBUF_SIZE];
	DWORD curByte = 0;
	while(*lpszBytes)
	{
		binBuffer[curByte++] = atoi(lpszBytes);
		lpszBytes = nextTok(lpszBytes,' ',false);
	}

	DWORD bw;

	// Figure out how many bytes we CAN write
	PeekNamedPipe(lpTP->hStdInR,szBuffer,OUTBUF_SIZE,&bw,NULL,NULL);

	bw = OUTBUF_SIZE - bw;
	if (bw > curByte)
		bw = curByte;
	
	if (!WriteFile(lpTP->hStdIn,binBuffer,bw,&bw,NULL))
		RETURNERROR(UNKNOWN);

	strcpy(data,"OK ");
	char *lpszNum = data+3;
	itoa(bw,lpszNum,10);
	lpszNum += strlen(lpszNum);
	*(lpszNum++) = ' ';
	itoa(curByte,lpszNum,10);
	return 3;
}


int __stdcall CloseInput(HWND mWnd, HWND aWnd, char *data, char *parms,
						 BOOL show, BOOL nopause)
{
	StayLoaded();

	char *lpszText = nextTok(data,' ',true);
	
	TEXTPROCESS *lpTP;
	if (!Hash_GetKeyCI(lpProcs,data,(void **)&lpTP))
		RETURNERROR(NONAME);

	if (!lpTP->hStdIn)
		RETURNERROR(INPUT);

	CloseHandle(lpTP->hStdInR);
	lpTP->hStdInR = NULL;
	if (!CloseHandle(lpTP->hStdIn))
	{
		lpTP->hStdIn = NULL;
		RETURNERROR(UNKNOWN);
	}
	lpTP->hStdIn = NULL;

	strcpy(data,"OK");
	return 3;
}

int __stdcall GetExitCode(HWND mWnd, HWND aWnd, char *data, char *parms,
						  BOOL show, BOOL nopause)
{
	StayLoaded();

	TEXTPROCESS *lpTP;
	if (!Hash_GetKeyCI(lpProcs,data,(void **)&lpTP))
		RETURNERROR(NONAME);

	DWORD termStat;
	GetExitCodeProcess(lpTP->hProcess,&termStat);
	if (termStat == STILL_ACTIVE)
		RETURNERROR(ALIVE);

	strcpy(data,"OK ");
	itoa(termStat,data+3,10);
	return 3;
}


int __stdcall SetProcess(HWND mWnd, HWND aWnd, char *data, char *parms,
						 BOOL show, BOOL nopause)
{
	StayLoaded();

	char *lpszFlag = nextTok(data,' ',true);

	TEXTPROCESS *lpTP;
	if (!Hash_GetKeyCI(lpProcs,data,(void **)&lpTP))
		RETURNERROR(NONAME);

	if (!stricmp(lpszFlag,"keepalive"))
		lpTP->fBlock |= TP_KEEPALIVE;
	else if (!stricmp(lpszFlag,"!keepalive"))
		lpTP->fBlock &= ~TP_KEEPALIVE;
	else
		RETURNERROR(SYNTAX);

	strcpy(data,"OK");
	return 3;
}

int __stdcall ReleaseProcess(HWND mWnd, HWND aWnd, char *data, char *parms,
							 BOOL show, BOOL nopause)
{
	StayLoaded();

	TEXTPROCESS *lpTP;
	if (!Hash_GetKeyCI(lpProcs,data,(void **)&lpTP))
		RETURNERROR(NONAME);

	DWORD termStat;
	GetExitCodeProcess(lpTP->hProcess,&termStat);

	Hash_DeleteKeyCI(lpProcs,data);


	if (lpTP->hStdIn)
	{
		CloseHandle(lpTP->hStdIn);
		CloseHandle(lpTP->hStdInR);
	}
	CloseHandle(lpTP->hStdOut);
	CloseHandle(lpTP->hStdErr);

	// Kill the process if it's still alive
	if ((termStat != STILL_ACTIVE) || (lpTP->fBlock & TP_KEEPALIVE))
	{
		CloseHandle(lpTP->hProcess);
		delete lpTP;
		strcpy(data,"OK released");
		return 3;
	}
	
	PostThreadMessage(lpTP->dwThreadId,WM_QUIT,0,0);

	// We'll give it a small chance to die on its own
	if (WaitForSingleObject(lpTP->hProcess,100) != WAIT_OBJECT_0)
		TerminateProcess(lpTP->hProcess,1);
	CloseHandle(lpTP->hProcess);
	delete lpTP;

	strcpy(data,"OK killed");
	
	return 3;
}

int __stdcall Remove(HWND mWnd, HWND aWnd, char *data, char *parms,
					 BOOL show, BOOL nopause)
{
	StayLoaded();

	FreeLibrary(Self);

	strcpy(data,"OK");
	return 3;
}

int __stdcall DLLInfo(HWND mWnd, HWND aWnd, char *data, char *parms,
					  BOOL show, BOOL nopause)
{
	strcpy(data,AUTHINFO);
	return 3;
}

#ifdef __cplusplus
}
#endif

